# coding=utf-8
""" Google Earth Engine Sentinel Collections """
from . import Collection, TODAY, Band
from .. import tools, cloud_mask
import ee

NUNBERS = [1, 2]
PROCESSES = ['TOA', 'SR']
IDS = [
    'COPERNICUS/S2',
    'COPERNICUS/S2_SR'
]


class Sentinel2(Collection):
    """ Sentinel 2 Collection """
    SHORTS = {
        'COPERNICUS/S2': 'S2TOA',
        'COPERNICUS/S2_SR': 'S2SR'
    }

    def __init__(self, process='TOA'):
        super(Sentinel2, self).__init__()
        if process not in PROCESSES:
            msg = '{} is not a valid process'
            raise ValueError(msg.format(process))

        if process == 'TOA':
            self._id = 'COPERNICUS/S2'
        else:
            self._id = 'COPERNICUS/S2_SR'

        self.number = 2
        self.spacecraft = 'SENTINEL2'
        self.process = process

        self.start_date = '2015-06-23'
        self.end_date = TODAY

        self.cloud_cover = 'CLOUD_COVERAGE_ASSESSMENT'
        self.algorithms = {
            'hollstein': cloud_mask.applyHollstein
        }

        if self.process == 'SR':
            self.algorithms['scl_masks'] = self.SclMasks
            self.common_masks = [self.scl, self.qa60]
        elif self.process == 'TOA':
            self.common_masks = [self.qa60]

        self._bands = self._make_bands()
        self.short_name = self.SHORTS.get(self._id)

    def scl(self, image, classes=('saturated', 'dark', 'shadow', 'cloud_low',
                        'cloud_medium', 'cloud_high', 'cirrus_cloud', 'snow'),
            renamed=False):
        for cls in classes:
            mask = ee.Image(self.SclMasks(image, renamed)).select(cls)
            if cls in ['saturated', 'dark', 'shadow', 'cloud_low',
                       'cloud_medium', 'cloud_high', 'cirrus_cloud', 'snow']:
                mask = mask.Not()
            image = image.updateMask(mask)
        return image

    def qa60(self, image, classes=('cloud', 'cirrus_cloud'), renamed=False):
        if renamed:
            band = 'qa60'
        else:
            band = 'QA60'

        return self.applyMask(image, band, classes, renamed)

    def _make_bands(self):
        band = [None]*30
        common = {'min':0, 'max': 10000, 'precision': 'uint16',
                  'reference': 'optical'}
        band[0] = Band('B1', 'aerosol', scale=60, **common)
        band[1] = Band('B2', 'blue', scale=10, **common)
        band[2] = Band('B3', 'green', scale=10, **common)
        band[3] = Band('B4', 'red', scale=10, **common)
        band[4] = Band('B5', 'red_edge_1', scale=20, **common)
        band[5] = Band('B6', 'red_edge_2', scale=20, **common)
        band[6] = Band('B7', 'red_edge_3', scale=20, **common)
        band[7] = Band('B8', 'nir', scale=10, **common)
        band[8] = Band('B8A', 'red_edge_4', scale=20, **common)
        band[9] = Band('B9', 'water_vapor', scale=60, **common)
        swir = Band('B11', 'swir', scale=20, **common)
        swir2 = Band('B12', 'swir2', scale=20, **common)
        qa10 = Band('QA10', 'qa10', scale=10, reference='bits',
                    precision='uint16')
        qa20 = Band('QA20', 'qa20', scale=20, reference='bits',
                    precision='uint32')
        qa60 = Band('QA60', 'qa60', scale=60, reference='bits',
                    bits={'10':{1:'cloud'}, '11':{1:'cirrus_cloud'}},
                    precision='uint16')

        if self.process in ['TOA']:
            band[10] = Band('B10', 'cirrus', scale=60, **common)
            band[11] = swir
            band[12] = swir2
            band[13] = qa10
            band[14] = qa20
            band[15] = qa60

        if self.process in ['SR']:
            band[10] = swir
            band[11] = swir2
            band[12] = Band('AOT', 'aerosol_thickness', 'uint16', 10,
                            0, 65535, 'optical')
            band[13] = Band('WVP', 'water_vapor_pressure', 'uint16', 10,
                            0, 65535, 'optical')
            band[14] = Band('SCL', 'scene_classification_map', 'uint8', 20,
                            1, 11, 'classification')
            band[15] = qa10
            band[16] = qa20
            band[17] = qa60

        return [b for b in band if b]

    def SclMasks(self, image, renamed=False):
        """ Decodify the SCL bands and create a mask for each category """
        if self.process == 'SR':
            if renamed:
                scl = image.select('scene_classification_map')
            else:
                scl = image.select('SCL')

            data = ee.Dictionary(self.SclData)

            def wrap(band_value, name):
                band_value = ee.Number.parse(band_value)
                name = ee.String(name)
                mask = scl.eq(band_value).rename(name)
                return mask

            newbands = ee.Dictionary(data.map(wrap))
            bandslist = tools.dictionary.extractList(newbands,
                                            [str(i) for i in range(1, 12)])

            image = tools.image.addMultiBands(bandslist)
        return image

    @property
    def SclData(self):
        data = None
        if self.process == 'SR':
            data = {
                1: 'saturated',
                2: 'dark',
                3: 'shadow',
                4: 'vegetation',
                5: 'bare_soil',
                6: 'water',
                7: 'cloud_low',
                8: 'cloud_medium',
                9: 'cloud_high',
                10: 'cirrus_cloud',
                11: 'snow'
            }
        return data

    @staticmethod
    def fromId(id):
        """ Create a Sentinel2 class from a GEE ID """
        if id == 'COPERNICUS/S2':
            return Sentinel2()
        elif id == 'COPERNICUS/S2_SR':
            return Sentinel2('SR')
        else:
            msg = '{} not recognized as a Sentinel 2 ID'
            raise ValueError(msg.format(id))

    # FACTORY
    @classmethod
    def Sentinel2TOA(cls):
        return cls('TOA')

    @classmethod
    def Sentienl2SR(cls):
        return cls('SR')


# coding=utf-8
""" Google Earth Engine MODIS Collections """

from . import Collection, TODAY, Band
from functools import partial

IDS = [
    'MODIS/006/MOD09GQ', 'MODIS/006/MYD09GQ',
    'MODIS/006/MOD09GA', 'MODIS/006/MYD09GA',
    'MODIS/006/MOD13Q1', 'MODIS/006/MYD13Q1'
]

START = {
    'MODIS/006/MOD09GQ': '2000-02-24',
    'MODIS/006/MYD09GQ': '2000-02-24',
    'MODIS/006/MOD09GA': '2000-02-24',
    'MODIS/006/MYD09GA': '2000-02-24',
    'MODIS/006/MOD13Q1': '2000-02-18',
    'MODIS/006/MYD13Q1': '2000-02-18',
}

END = {
    'MODIS/006/MOD09GQ': TODAY,
    'MODIS/006/MYD09GQ': TODAY,
    'MODIS/006/MOD09GA': TODAY,
    'MODIS/006/MYD09GA': TODAY,
    'MODIS/006/MOD13Q1': TODAY,
    'MODIS/006/MYD13Q1': TODAY,
}


class MODIS(Collection):
    """ MODIS Collections """

    SHORTS = {
        'MODIS/006/MOD09GQ': 'TERRA_SR_250_DAILY',
        'MODIS/006/MYD09GQ': 'AQUA_SR_250_DAILY',
        'MODIS/006/MOD09GA': 'TERRA_SR_1KM_DAILY',
        'MODIS/006/MYD09GA': 'AQUA_SR_1KM_DAILY',
        'MODIS/006/MOD13Q1': 'TERRA_IND_250_16DAYS',
        'MODIS/006/MYD13Q1': 'AQUA_IND_250_16DAYS'
    }

    def __init__(self, product_id):
        """ Initialize a MODIS collection with it's product id """
        super(MODIS, self).__init__()
        self.product_id = product_id
        self._id = self._make_id()
        self._bands = self._make_bands()

        # dates
        self.start_date = START[self._id]
        self.end_date = END[self._id]

        self.spacecraft = 'MODIS'

        self.cloud_cover = None

        self.short_name = self.SHORTS.get(self.id)

        if self._id in ['MODIS/006/MOD09GQ', 'MODIS/006/MYD09GQ']:
            self.common_masks = [self.qc250]
        if self._id in ['MODIS/006/MOD09GA', 'MODIS/006/MYD09GA']:
            self.common_masks = [self.state_1km]
        if self._id in ['MODIS/006/MOD13Q1', 'MODIS/006/MYD13Q1']:
            self.common_masks = [self.detailed_qa]

    def state_1km(self, image, classes=('cloud', 'shadow', 'snow',
                                        'average_cirrus', 'high_cirrus'), renamed=False):

        return self.applyMask(image, 'state_1km', classes, renamed)

    def qc250(self, image, classes=('B1_highest_quality', 'B2_highest_quality'),
              renamed=False):

        return self.applyPositiveMask(image, 'QC_250m', classes, renamed)

    def detailed_qa(self, image, classes=('cloud', 'shadow', 'snow'),
                    renamed=False):
        if renamed:
            band ='DetailedQA'
        else:
            band = 'detailed_qa'

        return self.applyMask(image, band, classes, renamed)

    def _make_bands(self):
        bands = [None]*30

        # Partial bands
        sur_refl_b01 = partial(Band, id='sur_refl_b01', name='red',
                               precision='int16', min=-100,
                               max=16000, reference='optical')

        sur_refl_b02 = partial(Band, id='sur_refl_b02', name='nir',
                               precision='int16', min=-100,
                               max=16000, reference='optical')

        num_observations = partial(Band, precision='int8', min=0, max=127,
                                   reference='classification')

        QC_250m = Band('QC_250m', 'QC_250m', 'uint16', 250, 0, 4096,
                       'bits', bits={
                            '4-7': {0: 'B1_highest_quality'},
                            '8-11': {0: 'B2_highest_quality'},
                            '12': {1: 'atmospheric_corrected'}
                        })

        obscov = partial(Band, precision='int8', min=0, max=100,
                         reference='classification')

        iobs_res = partial(Band, id='iobs_res', name='obs_number',
                           precision='uint8', min=0, max=254,
                           reference='classification')

        orbit_pnt = partial(Band, id='orbit_pnt', name='orbit_pointer',
                            precision='int8', min=0, max=15,
                            reference='classification')

        granule_pnt = partial(Band, id='granule_pnt', name='granule_pointer',
                              precision='uint8', min=0, max=254,
                              reference='classification')

        state_1km = Band('state_1km', 'state_1km', 'uint16', 1000, 0, 57335,
                         'bits', bits={
                            '0-1': {0: 'clear', 1:'cloud', 2:'mix'},
                            '2':   {1: 'shadow'},
                            '8-9': {1: 'small_cirrus', 2: 'average_cirrus',
                                    3: 'high_cirrus'},
                            '13':  {1: 'adjacent'},
                            '15':  {1: 'snow'}
                        })
        sezenith = Band('SensorZenith', 'sensor_zenith', 'int16', 1000, 0,
                        18000, 'classification')
        seazimuth = Band('SensorAzimuth', 'sensor_azimuth', 'int16', 1000,
                         -18000, 18000, 'classification')
        range_band = Band('Range', 'range', 'uint16', 1000, 27000, 65535,
                     'classification')
        sozenith = Band('SolarZenith', 'solar_zenith', 'int16', 1000, 0,
                        18000, 'classification')
        soazimuth = Band('SolarAzimuth', 'solar_azimuth', 'int16', 1000,
                         -18000, 18000, 'classification')
        gflags = Band('gflags', 'geolocation_flags', 'uint8', 1000, 0, 248,
                      'bits')

        sur_refl_b03 = partial(Band, id='sur_refl_b03', name='blue',
                               precision='int16', min=-100, max=16000,
                               reference='optical')

        sur_refl_b04 = partial(Band, id='sur_refl_b04', name='green',
                               precision='int16', min=-100, max=16000,
                               reference='optical')

        sur_refl_b05 = partial(Band, id='sur_refl_b05', name='swir3',
                               precision='int16', min=-100, max=16000,
                               reference='optical')

        sur_refl_b06 = partial(Band, id='sur_refl_b06', name='swir',
                               precision='int16', min=-100, max=16000,
                               reference='optical')

        sur_refl_b07 = partial(Band, id='sur_refl_b07', name='swir2',
                               precision='int16', min=-100, max=16000,
                               reference='optical')

        QC_500m = Band('QC_500m', 'QC_500m', 'uint32', 500, 0, 4294966019,
                       'bits', bits={
                            '2-5': {0: 'B1_highest_quality'},
                            '6-9': {0: 'B2_highest_quality'},
                            '10-13': {0: 'B3_highest_quality'},
                            '14-17': {0: 'B4_highest_quality'},
                            '18-21': {0: 'B5_highest_quality'},
                            '22-25': {0: 'B6_highest_quality'},
                            '26-29': {0: 'B7_highest_quality'},
                        })

        qscan = Band('q_scan', 'q_scan', 'uint8', 250, 0, 254, 'bits')

        NDVI = Band('NDVI', 'ndvi', 'int16', 250, -2000, 10000, 'classification')

        EVI = Band('EVI', 'evi', 'int16', 250, -2000, 10000, 'classification')

        DetailedQA = Band('DetailedQA', 'detailed_qa', 'uint16', 250, 0, 65534,
                          'bits', bits={
                            '0-1': {0: 'good_qa'},
                            '2-5': {0: 'highest_qa'},
                            '8':   {1: 'adjacent'},
                            '10':  {1: 'cloud'},
                            '14':  {1: 'snow'},
                            '15':  {1: 'shadow'}
                          })

        view_zenith = Band('ViewZenith', 'view_zenith', 'int16', 250, 0, 18000,
                           'classification')

        relative_azimuth = Band('RelativeAzimuth', 'relative_azimuth', 'int16',
                                250, -18000, 18000, 'classification')

        DayOfYear = Band('DayOfYear', 'day_of_year', 'int16', 250, 1, 366,
                         'classification')

        SummaryQA = Band('SummaryQA', 'summary_qa', 'int8', 250, 0, 3, 'bits',
                         bits={
                             '0-1': {0: 'clear', 1: 'marginal', 2: 'snow',
                                     3: 'cloud'}
                         })

        if self.product_id in ['MOD09GQ', 'MYD09GQ']:
            bands[0] = num_observations(id='num_observations',
                                        name='num_observations', scale=250)
            bands[1] = sur_refl_b01(scale=250)
            bands[2] = sur_refl_b02(scale=250)
            bands[3] = QC_250m
            bands[4] = obscov(id='obscov', name='observation_coverage', scale=250)
            bands[5] = iobs_res(scale=250)
            bands[6] = orbit_pnt(scale=250)
            bands[7] = granule_pnt(scale=250)

        if self.product_id in ['MOD09GA', 'MYD09GA']:
            bands[0] = num_observations(id='num_observations_1km', scale=1000,
                                        name='num_observations_1km')
            bands[1] = state_1km
            bands[2] = sezenith
            bands[3] = seazimuth
            bands[4] = range_band
            bands[5] = sozenith
            bands[6] = soazimuth
            bands[7] = gflags
            bands[8] = orbit_pnt(scale=500)
            bands[9] = granule_pnt(scale=500)
            bands[10] = num_observations(id='num_observations_500m', scale=500,
                                         name='num_observations_500m')
            bands[11] = sur_refl_b01(scale=500)
            bands[12] = sur_refl_b02(scale=500)
            bands[13] = sur_refl_b03(scale=500)
            bands[14] = sur_refl_b04(scale=500)
            bands[15] = sur_refl_b05(scale=500)
            bands[16] = sur_refl_b06(scale=500)
            bands[17] = sur_refl_b07(scale=500)
            bands[18] = QC_500m
            bands[19] = obscov(id='obscov_500m', scale=500,
                               name='observation_coverage_500m')
            bands[20] = iobs_res(scale=500)
            bands[21] = qscan

        if self.product_id in ['MOD13Q1', 'MYD13Q1']:
            bands[0] = NDVI
            bands[1] = EVI
            bands[2] = DetailedQA
            bands[3] = sur_refl_b01(scale=250)
            bands[4] = sur_refl_b02(scale=250)
            bands[5] = sur_refl_b03(scale=250)
            bands[6] = sur_refl_b07(scale=250)
            bands[7] = view_zenith
            bands[8] = sozenith
            bands[9] = relative_azimuth
            bands[10] = DayOfYear
            bands[11] = SummaryQA

        return [b for b in bands if b]

    def _make_id(self):
        return 'MODIS/006/{}'.format(self.product_id)

    @staticmethod
    def fromId(id):
        """ Make a MODIS collection from its ID """
        def error():
            msg = 'Collection {} not available'
            raise ValueError(msg.format(id))

        if id not in IDS: error()

        splitted = id.split('/')
        prod = splitted[2]

        return MODIS(prod)

    @classmethod
    def MOD09GQ(cls):
        return cls(product_id='MOD09GQ')

    @classmethod
    def MYD09GQ(cls):
        return cls(product_id='MYD09GQ')

    @classmethod
    def MOD09GA(cls):
        return cls(product_id='MOD09GA')

    @classmethod
    def MYD09GA(cls):
        return cls(product_id='MYD09GA')

    @classmethod
    def MOD13Q1(cls):
        return cls(product_id='MOD13Q1')

    @classmethod
    def MYD13Q1(cls):
        return cls(product_id='MYD13Q1')



# coding=utf-8
""" Google Earth Engine Landsat Collections """
from . import Collection, TODAY, Band
from .. import tools
from .. import algorithms as module_alg
import ee

NUMBERS = [1, 2, 3, 4, 5, 7, 8]
PROCESSES = ['RAW', 'TOA', 'SR']
SENSORS = ['TM', 'MSS', 'ETM', 'OLI']
TIERS = [1, 2]
NUMBER_ID = {
    1: {'MSS': 'LM01'},
    2: {'MSS': 'LM02'},
    3: {'MSS': 'LM03'},
    4: {'TM':'LT04', 'MSS':'LM04'},
    5: {'TM':'LT05', 'MSS':'LM05'},
    7: {'ETM':'LE07'},
    8: {'OLI':'LC08'}
}
TIER_ID = {1: 'T1', 2: 'T2'}
PROCESS_ID = {'RAW': '', 'SR': '_SR', 'TOA': '_TOA'}
START = {1: '1972-07-23', 2: '1975-01-22', 3: '1978-03-05',
         4: '1982-07-16', 5: '1984-01-01', 7: '1999-01-01',
         8: '2013-04-11'}
END = {1: '1978-01-07', 2: '1982-02-26', 3: '1983-03-31',
       4: '1993-12-14', 5: '2012-05-05', 7: TODAY, 8: TODAY}
DEFAULTS = {
    1: {'process': 'RAW', 'sensor':'MSS'},
    2: {'process': 'RAW', 'sensor':'MSS'},
    3: {'process': 'RAW', 'sensor':'MSS'},
    4: {'process': 'SR', 'sensor':'TM'},
    5: {'process': 'SR', 'sensor':'TM'},
    7: {'process': 'SR', 'sensor':'ETM'},
    8: {'process': 'SR', 'sensor':'OLI'},
}
NUMBER_PROCESS = {
    1: ['RAW'], 2: ['RAW'], 3: ['RAW'],
    4: ['RAW', 'TOA', 'SR'], 5: ['RAW', 'TOA', 'SR'],
    7: ['RAW', 'TOA', 'SR'], 8: ['RAW', 'TOA', 'SR']
}
NUMBER_SENSOR = {k: list(v.keys()) for k, v in NUMBER_ID.items()}
IDS = [
    'LANDSAT/LM01/C01/T1',
    'LANDSAT/LM01/C01/T2',
    'LANDSAT/LM02/C01/T1',
    'LANDSAT/LM02/C01/T2',
    'LANDSAT/LM03/C01/T1',
    'LANDSAT/LM03/C01/T2',
    'LANDSAT/LM04/C01/T1',
    'LANDSAT/LM04/C01/T2',
    'LANDSAT/LM05/C01/T1',
    'LANDSAT/LM05/C01/T2',
    'LANDSAT/LT04/C01/T1', 'LANDSAT/LT04/C01/T1_TOA', 'LANDSAT/LT04/C01/T1_SR',
    'LANDSAT/LT04/C01/T2', 'LANDSAT/LT04/C01/T2_TOA', 'LANDSAT/LT04/C01/T2_SR',
    'LANDSAT/LT05/C01/T1', 'LANDSAT/LT05/C01/T1_TOA', 'LANDSAT/LT05/C01/T1_SR',
    'LANDSAT/LT05/C01/T2', 'LANDSAT/LT05/C01/T2_TOA', 'LANDSAT/LT05/C01/T2_SR',
    'LANDSAT/LE07/C01/T1', 'LANDSAT/LE07/C01/T1_TOA', 'LANDSAT/LE07/C01/T1_SR',
    'LANDSAT/LE07/C01/T2', 'LANDSAT/LE07/C01/T2_TOA', 'LANDSAT/LE07/C01/T2_SR',
    'LANDSAT/LC08/C01/T1', 'LANDSAT/LC08/C01/T1_TOA', 'LANDSAT/LC08/C01/T1_SR',
    'LANDSAT/LC08/C01/T2', 'LANDSAT/LC08/C01/T2_TOA', 'LANDSAT/LC08/C01/T2_SR',
]


class Landsat(Collection):
    """  Landsat Collection """
    SHORTS = {
        'LANDSAT/LM01/C01/T1': 'L1', 'LANDSAT/LM01/C01/T2': 'L1T2',
        'LANDSAT/LM02/C01/T1': 'L2', 'LANDSAT/LM02/C01/T2': 'L2T2',
        'LANDSAT/LM03/C01/T1': 'L3', 'LANDSAT/LM03/C01/T2': 'L3T2',
        'LANDSAT/LM04/C01/T1': 'L4MSS', 'LANDSAT/LM04/C01/T2': 'L4MSST2',
        'LANDSAT/LM05/C01/T1': 'L5MSS', 'LANDSAT/LM05/C01/T2': 'L5MSST2',
        'LANDSAT/LT04/C01/T1': 'L4RAW', 'LANDSAT/LT04/C01/T1_TOA': 'L4TOA', 'LANDSAT/LT04/C01/T1_SR':'L4SR',
        'LANDSAT/LT04/C01/T2': 'L4RAWT2', 'LANDSAT/LT04/C01/T2_TOA': 'L4TOAT2', 'LANDSAT/LT04/C01/T2_SR': 'L4SRT2',
        'LANDSAT/LT05/C01/T1': 'L5RAW', 'LANDSAT/LT05/C01/T1_TOA': 'L5TOA', 'LANDSAT/LT05/C01/T1_SR': 'L5SR',
        'LANDSAT/LT05/C01/T2': 'L5RAWT2', 'LANDSAT/LT05/C01/T2_TOA': 'L5TOAT2', 'LANDSAT/LT05/C01/T2_SR': 'L5SRT2',
        'LANDSAT/LE07/C01/T1': 'L7RAW', 'LANDSAT/LE07/C01/T1_TOA': 'L7TOA', 'LANDSAT/LE07/C01/T1_SR': 'L7SR',
        'LANDSAT/LE07/C01/T2': 'L7RAWT2', 'LANDSAT/LE07/C01/T2_TOA': 'L7TOAT2', 'LANDSAT/LE07/C01/T2_SR': 'L7SRT2',
        'LANDSAT/LC08/C01/T1': 'L8RAW', 'LANDSAT/LC08/C01/T1_TOA': 'L8TOA', 'LANDSAT/LC08/C01/T1_SR': 'L8SR',
        'LANDSAT/LC08/C01/T2': 'L8RAWT2', 'LANDSAT/LC08/C01/T2_TOA': 'L8TOAT2', 'LANDSAT/LC08/C01/T2_SR': 'L8SRT2',
    }

    def __init__(self, number, process=None, sensor=None, tier=1):
        """ Landsat Collection

        :param number: One of 1, 2, 3, 4, 5, 7, 8
        :param process: One of 'RAW', 'TOA', 'SR'
        :param sensor: One of 'TM', 'MSS', 'ETM', 'OLI'
        :param tier: One of 1, 2
        """
        super(Landsat, self).__init__()
        self.number = number
        self.process = process if process else DEFAULTS[number]['process']
        self.sensor = sensor if sensor else DEFAULTS[number]['sensor']
        self.tier = tier

        # CHECK AVAILABLE
        if self.process not in NUMBER_PROCESS[number]:
            msg = "Process {} not available for Landsat {}"
            raise ValueError(msg.format(self.process, number))

        if self.sensor not in NUMBER_SENSOR[number]:
            msg = "Sensor {} not available for Landsat {}"
            raise ValueError(msg.format(self.sensor, number))

        # Catch L4 MSS TOA or SR
        if self.sensor == 'MSS' and self.process in ['SR', 'TOA']:
            msg = 'Process {} not available for sensor {}'
            self.process = 'RAW'

        # Landsat common properties
        self.cloud_cover = 'CLOUD_COVER'
        self.spacecraft = 'LANDSAT'

        # set algorithms
        self.algorithms['brdf'] = self.brdf
        self.algorithms['harmonize'] = self.harmonize
        self.algorithms['rescale'] = self.rescaleAll

        # dates
        self.start_date = START[self.number]
        self.end_date = END[self.number]

        # ID
        self._id = self._make_id()

        # BANDS
        self._bands = self._make_bands()

        self.short_name = self.SHORTS.get(self.id)

        # COMMON MASKS
        if self.sensor == 'MSS':
            self.common_masks = [self.bqa_mss]
        if self.number in [4,5,7]:
            if self.process == 'SR':
                self.common_masks = [self.pixel_qa, self.sr_cloud_qa]
            elif self.process in ['RAW', 'TOA']:
                self.common_masks = [self.bqa]
        if self.number == 8:
            if self.process == 'SR':
                self.common_masks = [self.pixel_qa]
            elif self.process in ['TOA', 'RAW']:
                self.common_masks = [self.bqa]

    @staticmethod
    def fromId(id):
        """ Create a Landsat class from a GEE ID """
        if id not in IDS:
            msg = 'Collection {} not available'
            raise ValueError(msg.format(id))

        sensors = {'LT': 'TM', 'LM': 'MSS', 'LE': 'ETM', 'LC': 'OLI'}
        # decompose id
        parts = id.split('/')
        shortid = parts[1]
        sensor = sensors[shortid[0:2]]
        number = int(shortid[2:4])
        process_tier = parts[3]
        process_tier_split = process_tier.split('_')

        if len(process_tier_split) == 1:
            process = 'RAW'
            tier = process_tier_split[0]
        else:
            process = process_tier_split[1]
            tier = process_tier_split[0]

        return Landsat(number, process, sensor, int(tier[1]))

    def _make_id(self):
        """ Make ID from parsed data """
        return 'LANDSAT/{}/C01/{}{}'.format(
            NUMBER_ID[self.number][self.sensor],
            TIER_ID[self.tier],
            PROCESS_ID[self.process])

    def _make_bands(self):
        """ Band's name relation. Return a list of bands (Band instances) """
        band = [None]*20
        number = self.number
        process = self.process
        sensor = self.sensor

        max_raw_8 = 65535
        max_raw = 255
        max_toa_optical = 1
        max_toa_thermal = 1000
        max_sr = 10000

        # 457 SR
        atm_op = Band('sr_atmos_opacity', 'atmos_opacity', 'int16', 30,
                      -32768, 32767, reference='classification')
        cloud_qa_bits = {'0': {1:'ddv'},
                         '1': {1:'cloud'},
                         '2': {1:'shadow'},
                         '3': {1:'adjacent'},
                         '4': {1:'snow'},
                         '5': {1:'water'}}
        sr_cloud_qa = Band('sr_cloud_qa', 'cloud_qa', 'uint8', 30,
                           reference='bits', bits=cloud_qa_bits)

        pixel_qa = Band('pixel_qa', 'pixel_qa', 'uint16', 30,
                        reference='bits')

        radsat_qa_457 = Band('radsat_qa', 'radsat_qa', 'uint8', 30,
                             reference='bits',
                             bits={
                                 1: {1:'B1_saturated'},
                                 2: {1:'B2_saturated'},
                                 3: {1:'B3_saturated'},
                                 4: {1:'B4_saturated'},
                                 5: {1:'B5_saturated'},
                                 6: {1:'B6_saturated'},
                                 7: {1:'B7_saturated'},
                             })

        radsat_qa_8 = Band('radsat_qa', 'radsat_qa', 'uint16', 30,
                           reference='bits',
                           bits={
                               1: {1:'B1_saturated'},
                               2: {1:'B2_saturated'},
                               3: {1:'B3_saturated'},
                               4: {1:'B4_saturated'},
                               5: {1:'B5_saturated'},
                               6: {1:'B6_saturated'},
                               7: {1:'B7_saturated'},
                               9: {1:'B9_saturated'},
                               10: {1:'B10_saturated'},
                               11: {1:'B11_saturated'},
                           })

        bqa = Band('BQA', 'bqa', 'uint16', 30, reference='bits',
                   bits={'4': {1:'cloud'},
                         '5-6': {3:'high_confidence_cloud'},
                         '7-8': {3:'shadow'},
                         '9-10': {3:'snow'}})

        sr_aerosol = Band('sr_aerosol', 'sr_aerosol', 'uint8', 30,
                          reference='bits',
                          bits={
                              '3': {1: 'water'},
                              '6-7': {0: 'climatology',
                                      1: 'low',
                                      2:'medium',
                                      3: 'high'}
                          })

        if sensor in ['MSS']:
            green = Band('B1', 'green', 'int8', 60, 0, 255, 'optical')
            red = Band('B2', 'red', 'int8', 60, 0, 255, 'optical')
            nir = Band('B3', 'nir', 'int8', 60, 0, 255, 'optical')
            nir2 = Band('B4', 'nir2', 'int8', 30, 0, 255, 'optical')
            qa = Band('BQA', 'bqa', 'int16', 60, reference='bits',
                       bits={'4': {1: 'cloud'}})
            if number in [1, 2, 3]:
                green.id = 'B4'
                red.id = 'B5'
                nir.id = 'B6'
                nir2.id = 'B7'
            band[0] = green
            band[1] = red
            band[2] = nir
            band[3] = nir2
            band[4] = qa

        # Opticals
        if number in [4, 5, 7]:
            common = dict(scale=30, reference='optical')
            band[0] = Band('B1', 'blue', **common)
            band[1] = Band('B2', 'green', **common)
            band[2] = Band('B3', 'red', **common)
            band[3] = Band('B4', 'nir', **common)
            band[4] = Band('B5', 'swir', **common)
            band[5] = Band('B6', 'thermal', scale=30, reference='thermal')
            band[6] = Band('B7', 'swir2', **common)
            if process in ['SR']:
                # set precision for SR
                for b in band:
                    if not b: continue
                    b.precision = 'int16'
                    b.min = 0
                    b.max = max_sr
                band[7] = atm_op
                band[8] = sr_cloud_qa
                pixel_qa.bits = {'1': {1:'clear'}, '2': {1:'water'},
                                 '3': {1:'shadow'}, '4': {1:'snow'},
                                 '5': {1:'cloud'},
                                 '6-7':{3:'high_confidence_cloud'}}
                band[9] = pixel_qa
                band[10] = radsat_qa_457
            if process in ['TOA']:
                for b in band:
                    if not b: continue
                    b.precision = 'float'
                    b.min = 0
                    b.max = max_toa_optical
            if process in ['RAW']:
                for b in band:
                    if not b: continue
                    b.precision = 'uint8'
                    b.min = 0
                    b.max = max_raw

        # BQA
        if number in [4, 5] and process in ['RAW', 'TOA']:
            band[7] = bqa

        if number in [7] and process in ['RAW', 'TOA']:
            band[9] = bqa
            # change swir2 from position 6 to 7
            band[7] = band[6]

        # B6_VCID_1/2 bands. Overwrites band 5 set before
        if number in [7] and process in ['TOA']:
            band[5] = Band('B6_VCID_1', 'thermal', 'float', 30, 0,
                           max_toa_thermal, reference='thermal')
            band[6] = Band('B6_VCID_2', 'thermal2', 'float', 30, 0,
                           max_toa_thermal, reference='thermal')
            # set pan band
            band[8] = Band('B8', 'pan', 'float', 15, 0, max_toa_optical,
                           reference='optical')

        if number in [7] and process in ['RAW']:
            band[5] = Band('B6_VCID_1', 'thermal', 'uint8', 30, 0, max_raw,
                           reference='thermal')
            band[6] = Band('B6_VCID_2', 'thermal2', 'uint8', 30, 0,
                           max_raw, reference='thermal')
            # set pan band
            band[8] = Band('B8', 'pan', 'uint8', 15, 0, max_raw,
                           reference='optical')

        if self.number == 8:
            common = dict(scale=30)
            thermal = Band('B10', 'thermal', reference='thermal', **common)
            thermal2 = Band('B11', 'thermal2', reference='thermal',**common)
            pan = Band('B8', 'pan', reference='optical', scale=15)
            cirrus = Band('B9', 'cirrus', reference='optical', scale=15)

            band[0] = Band('B1', 'aerosol', reference='optical', **common)
            band[1] = Band('B2', 'blue', reference='optical', **common)
            band[2] = Band('B3', 'green', reference='optical', **common)
            band[3] = Band('B4', 'red', reference='optical',**common)
            band[4] = Band('B5', 'nir', reference='optical',**common)
            band[5] = Band('B6', 'swir', reference='optical',**common)
            band[6] = Band('B7', 'swir2', reference='optical', **common)
            if process in ['SR']:
                band[7] = thermal
                band[8] = thermal2
                for b in band:
                    if not b: continue
                    b.precision = 'int16'
                    b.min = 0
                    b.max = max_sr
                band[9] = sr_aerosol
                pixel_qa.bits = {'1': {1:'clear'}, '2': {1:'water'},
                                 '3': {1:'shadow'}, '4': {1:'snow'},
                                 '5': {1:'cloud'},
                                 '6-7':{3:'high_confidence_cloud'},
                                 '8-9':{3:'high_confidence_cirrus'}
                                 }
                band[10] = pixel_qa
                band[11] = radsat_qa_8
            if process in ['TOA']:
                band[7] = pan
                band[8] = cirrus
                band[9] = thermal
                band[10] = thermal2
                for b in band:
                    if not b: continue
                    b.precision = 'float'
                    b.min = 0
                    if b.reference == 'optical':
                        b.max = max_toa_optical
                    elif b.reference == 'thermal':
                        b.max = max_toa_thermal
                band[11] = bqa
            if process in ['RAW']:
                band[7] = pan
                band[8] = cirrus
                band[9] = thermal
                band[10] = thermal2
                for b in band:
                    if not b: continue
                    b.precision = 'uint16'
                    b.min = 0
                    b.max = max_raw_8
                band[11] = bqa

        return [b for b in band if b]

    def bqa_mss(self, image, classes=('cloud',), renamed=False):
        if renamed:
            band = 'bqa'
        else:
            band = 'BQA'

        return self.applyMask(image, band, classes, renamed)

    def sr_cloud_qa(self, image, classes=('cloud', 'shadow', 'snow'), renamed=False):
        if renamed:
            band = 'cloud_qa'
        else:
            band = 'sr_cloud_qa'

        return self.applyMask(image, band, classes, renamed)

    def pixel_qa(self, image, classes=('shadow', 'cloud', 'snow'), renamed=False):
        for clas in classes:
            if clas in ['clear', 'water']:
                image = self.applyPositiveMask(image, 'pixel_qa', [clas], renamed)
            else:
                image = self.applyMask(image, 'pixel_qa', [clas], renamed)
        return image

    def bqa(self, image, classes=('cloud', 'shadow', 'snow'), renamed=False):
        if renamed:
            band = 'bqa'
        else:
            band = 'BQA'
        return self.applyMask(image, band, classes, renamed)

    def harmonize(self, image, renamed=False):
        """ HARMONIZATION """
        options = ['blue', 'green', 'red', 'nir', 'swir', 'swir2']
        if self.number == 8:
            if renamed:
                bands = {band.name:band.name for band in self.bands if band.name in options}
            else:
                bands = {band.name:band.id for band in self.bands if band.name in options}

            max_value = max(
                [band.max for band in self.bands if band.name in options])
            harmonized = module_alg.Landsat.harmonization(
                image, max_value=max_value, **bands)
        else:
            harmonized = image

        return harmonized

    def brdf(self, image, renamed=False):
        """ BRDF Correction """
        # BRDF
        bands = []
        all_present = True
        for band in ['red', 'green', 'blue', 'nir', 'swir', 'swir2']:
            if not renamed:
                bid = self.getBand(band, 'name').id
            else:
                bid = self.getBand(band, 'name').name
            if not bid:
                all_present = False
                break
            bands.append(bid)

        if all_present:
            return module_alg.Landsat.brdfCorrect(image, *bands)
        else:
            return image

    def _rescale(self, band_type, image, number, process, drop=False,
                 renamed=False):
        # Create comparative collection
        bands = ee.Dictionary(self.bands)

        proxy = Landsat(number, process)

        if band_type == 'thermal':
            this_band = list(self.thermalBands.keys())
            proxy_band = list(proxy.thermalBands.keys())

        if band_type == 'optical':
            this_band = list(self.opticalBands.keys())
            proxy_band = list(proxy.opticalBands.keys())

        # get common bands for thermal (thermal1, etc..)
        # common_bands = ee.List([band for band in this_band if band in
        # proxy_band])
        common_bands = tools.ee_list.intersection(ee.List(this_band),
                                                  ee.List(proxy_band))

        ranges_this = ee.Dictionary(self.ranges)
        ranges_proxy = ee.Dictionary(proxy.ranges)

        def iteration(band, ini):
            ini = ee.Image(ini)
            band = ee.String(band)
            ranges_this_band = ee.Dictionary(ranges_this.get(band))
            ranges_proxy_band = ee.Dictionary(ranges_proxy.get(band))
            min_this = ee.Number(ranges_this_band.get('min'))
            min_proxy = ee.Number(ranges_proxy_band.get('min'))
            max_this = ee.Number(ranges_this_band.get('max'))
            max_proxy = ee.Number(ranges_proxy_band.get('max'))
            if not renamed:
                band = ee.String(bands.get(band))
            return tools.image.parametrize(ini,
                                           (min_this, max_this),
                                           (min_proxy, max_proxy),
                                           bands=[band])

        final = ee.Image(common_bands.iterate(iteration, image))
        if drop:
            if not renamed:
                common_bands = tools.dictionary.extractList(self.bands,
                                                            common_bands)
            final = final.select(common_bands)

        return final

    def rescaleThermal(self, image, number, process, drop=False,
                       renamed=False):
        """ Re-scale only the thermal bands of an image to match the band
        from the given number and process """
        return self._rescale('thermal', image, number, process, drop, renamed)

    def rescaleOptical(self, image, number, process, drop=False,
                       renamed=False):
        """ Re-scale only the optical bands of an image to match the band
        from the given number and process """
        return self._rescale('optical', image, number, process, drop, renamed)

    def rescaleAll(self, image, number, process, drop=False, renamed=False):
        """ Re-scale only the optical and thermal bands of an image to match
        the band from the given number and process """
        if drop:
            optical = self._rescale('optical', image, number,
                                    process, True, renamed)

            thermal = self._rescale('thermal', image, number,
                                process, True, renamed)

            all = optical.addBands(thermal)
        else:
            optical = self._rescale('optical', image, number,
                                    process, False, renamed)

            all = self._rescale('thermal', optical, number,
                                process, False, renamed)

        return all

    # FACTORY
    @classmethod
    def Landsat1(cls, tier=1):
        return cls(1, tier=tier)

    @classmethod
    def Landsat2(cls, tier=1):
        return cls(2, tier=tier)

    @classmethod
    def Landsat3(cls, tier=1):
        return cls(3, tier=tier)

    @classmethod
    def Landsat4SR(cls, tier=1):
        return cls(4, sensor='TM', process='SR', tier=tier)

    @classmethod
    def Landsat4TOA(cls, tier=1):
        return cls(4, sensor='TM', process='TOA', tier=tier)

    @classmethod
    def Landsat5SR(cls, tier=1):
        return cls(5, process='SR', tier=tier)

    @classmethod
    def Landsat5TOA(cls, tier=1):
        return cls(5, process='TOA', tier=tier)

    @classmethod
    def Landsat7SR(cls, tier=1):
        return cls(7, process='SR', tier=tier)

    @classmethod
    def Landsat7TOA(cls, tier=1):
        return cls(7, process='TOA', tier=tier)

    @classmethod
    def Landsat8SR(cls, tier=1):
        return cls(8, process='SR', tier=tier)

    @classmethod
    def Landsat8TOA(cls, tier=1):
        return cls(8, process='TOA', tier=tier)



# coding=utf-8
""" Group of collections """
from . import getCommonBands, rescale
from datetime import date

class CollectionGroup(object):
    def __init__(self, *args):
        self.collections = args

    @property
    def ids(self):
        return [col.id for col in self.collections]

    def commonBands(self, reference='all', match='id'):
        """ Get a list of the bands that exist in all collections """
        return getCommonBands(*self.collections, reference=reference,
                              match=match)

    def scales(self):
        """ Get the minimum scale value that takes evey common band """
        scales = {}
        common = self.commonBands()
        for band in common:
            band_scales = []
            for collection in self.collections:
                scale = collection.scales()[band]
                band_scales.append(scale)
            scales[band] = {'min': min(band_scales),
                            'max': max(band_scales)}

        return scales

    def start_date(self):
        """ Get the absolute start date for the group """
        starts = [date.fromisoformat(c.start_date) for c in self.collections]
        return min(starts).isoformat()

    def end_date(self):
        """ Get the absolute end date for the group """
        starts = [date.fromisoformat(c.end_date) for c in self.collections]
        return max(starts).isoformat()



# coding=utf-8
""" Google Earth Engine collections """
import ee
from .. import indices, bitreader
from datetime import date
from .. import tools

TODAY = date.today().isoformat()


# HELPERS
def allequal(iterable):
    """ Check if all elements inside an iterable are equal """
    first = iterable[0]
    rest = iterable[1:]
    for item in rest:
        if item == first: continue
        else: return False
    return True


def convertPrecision(image, precision):
    """ Convert data type precisions """
    TYPES = ee.Dictionary({'float': image.toFloat(),
                           'double': image.toDouble(),
                           'int8': image.toInt8(),
                           'uint8': image.toUint8(),
                           'uint16': image.toUint16(),
                           'int16': image.toInt16(),
                           'uint32': image.toUint32(),
                           'int32': image.toInt32(),
                           'int64': image.toInt64()
                           })
    return ee.Image(TYPES.get(precision))


def convertPrecisions(image, precision_dict):
    precisions = ee.Dictionary(precision_dict)
    bands = ee.List(precisions.keys())
    def iteration(band, ini):
        ini = ee.Image(ini)
        imgband = ini.select([band])
        precision = ee.String(precisions.get(band))
        newband = convertPrecision(imgband, precision)
        return tools.image.replace(ini, band, newband)

    return ee.Image(bands.iterate(iteration, image))


def info(collection, renamed=False):
    """ Get the information for the parsed collection

    :param collection: the collection to get the information from
    :type collection: Collection
    :rtype: dict
    """
    data = dict()
    data['spacecraft'] = collection.spacecraft
    data['id'] = collection.id
    data['bands'] = [b.id for b in collection.bands]
    data['band_names'] = [b.name for b in collection.bands]
    data['ranges'] = collection.ranges(renamed=renamed)
    data['scales'] = collection.scales(renamed=renamed)
    data['start_date'] = collection.start_date
    data['end_date'] = collection.end_date
    data['algorithms'] = collection.algorithms
    if not renamed:
        data['thermal_bands'] = [b.id for b in collection.thermalBands]
        data['bit_bands'] = [b.id for b in collection.bitBands]
        data['optical_bands'] = [b.id for b in collection.opticalBands]
        data['classification_bands'] = [b.id for b in collection.classificationBands]
    else:
        data['thermal_bands'] = [b.name for b in collection.thermalBands]
        data['bit_bands'] = [b.name for b in collection.bitBands]
        data['optical_bands'] = [b.name for b in collection.opticalBands]
        data['classification_bands'] = [b.name for b in collection.classificationBands]

    data['cloud_cover'] = collection.cloud_cover
    data['ee_collection'] = collection.collection
    data['indices'] = collection.indices

    if (collection.spacecraft == 'LANDSAT'):
        data['sensor'] = collection.sensor
        data['process'] = collection.process
        data['tier'] = collection.tier
        data['number'] = collection.number

    if (collection.spacecraft == 'SENTINEL'):
        data['number'] = collection.number
        if collection.number == 2:
            data['process'] = collection.process
            if collection.process == 'SR':
                data['SCL'] = collection.SCL_data
    return data


def infoEE(collection):
    """ Return an information ee.Dictionary """
    information = info(collection)
    information.pop('algorithms')
    information.pop('ee_collection')
    information.pop('indices')
    information.pop('visualization')

    return ee.Dictionary(information)


class Band(object):
    """ Bands """
    accepted_precisions = ('float', 'double', 'int8', 'uint8', 'uint16',
                           'int16', 'uint32', 'int32', 'int64', None)

    def __init__(self, id, name, precision=None, scale=None,
                 min=None, max=None, reference=None, bits=None):
        self.id = id
        self.name = name
        self.precision = precision
        self.max = max
        self.min = min
        self.scale = scale
        self.reference = reference # thermal, optical or quality
        self.bits = bits

    @property
    def precision(self):
        return self._precision

    @precision.setter
    def precision(self, value):
        if value in self.accepted_precisions:
            self._precision = value
        else:
            msg = 'Presicion must be one of {}'
            raise ValueError(msg.format(self.accepted_precisions))


class Collection(object):
    """ Parent class for common operations """
    SHORTS = {}

    def __init__(self, **kwargs):
        self._id = None
        self._bands = None
        self.spacecraft = kwargs.get('spacecraft', None)
        self.start_date = kwargs.get('start_date', None)
        self.end_date = kwargs.get('end_date', None)
        self.cloud_cover = kwargs.get('cloud_cover', None)
        self.algorithms = kwargs.get('algorithms', {})
        self.common_masks = [lambda i: i]
        self.short_name = self.SHORTS.get(self.id)

    @staticmethod
    def fromId(id):
        return None

    @classmethod
    def fromShortName(cls, short_name):
        inverse = {v: k for k, v in cls.SHORTS.items()}
        if short_name not in inverse:
            raise ValueError("{} not recognized as a valid short name".format(short_name))
        else:
            cid = inverse[short_name]
            return cls.fromId(cid)

    @property
    def id(self):
        return self._id

    @property
    def bands(self):
        return self._bands

    @property
    def collection(self):
        """ Google Earth Engine Original Image Collection """
        return ee.ImageCollection(self.id)

    @property
    def opticalBands(self):
        return [band for band in self.bands if band.reference == 'optical']

    @property
    def thermalBands(self):
        return [band for band in self.bands if band.reference == 'thermal']

    @property
    def bitBands(self):
        return [band for band in self.bands if band.reference == 'bits']

    @property
    def classificationBands(self):
        return [band for band in self.bands if band.reference == 'classification']

    def bandNames(self, reference='all', renamed=False):
        """ List of band names """
        if reference == 'all':
            if not renamed:
                bands = [band.id for band in self.bands]
            else:
                bands = [band.name for band in self.bands]
        else:
            if not renamed:
                bands = [band.id for band in self.bands if band.reference == reference]
            else:
                bands = [band.name for band in self.bands if band.reference == reference]
        return bands

    def precisions(self, reference='all', renamed=False):
        precisions_dict = {}
        for band in self.bands:
            if reference != 'all' and reference != band.reference:
                continue
            if not renamed:
                name = band.id
            else:
                name = band.name

            precisions_dict[name] = band.precision

        return precisions_dict

    def ranges(self, reference='all', renamed=False):
        ranges_dict = {}
        for band in self.bands:
            if reference != 'all' and reference != band.reference:
                continue
            if not renamed:
                name = band.id
            else:
                name = band.name

            ranges_dict[name] = {'min': band.min, 'max': band.max}

        return ranges_dict

    def scales(self, reference='all', renamed=False):
        scales_dict = {}
        for band in self.bands:
            if reference != 'all' and reference != band.reference:
                continue
            if not renamed:
                name = band.id
            else:
                name = band.name

            scales_dict[name] = band.scale

        return scales_dict

    def bitOptions(self, renamed=False):
        options = {}
        for band in self.bands:
            if band.reference == 'bits' and band.bits:
                name = band.name if renamed else band.id
                opt = []
                for allclss in band.bits.values():
                    for _, clss in allclss.items():
                        opt.append(clss)
                options[name] = opt
        return options

    def getMask(self, image, mask_band, classes='all', renamed=False):
        """ Get a mask image """
        bandnames = self.bandNames(renamed=renamed)

        # SCL Mask
        if self.id == 'COPERNICUS/S2_SR' and mask_band == 'SCL':
            if classes == 'all':
                classes = ['dark','shadow', 'vegetation','bare_soil', 'water',
                'cloud_low', 'cloud_medium', 'cloud_high', 'cirrus_cloud',
                'snow']
            return ee.Image(self.SclMasks(image)).select(classes)

        # CHECKS
        if mask_band not in bandnames:
            msg = 'band {} not present in bands {}'
            raise ValueError(msg.format(mask_band, bandnames))
        options = self.bitOptions(renamed)[mask_band]
        if classes != 'all' and classes != ['all']:
            if not isinstance(classes, (list, tuple)):
                msg = 'classes param must be "all" or a list of classes'
                raise ValueError(msg)
            for clas in classes:
                if clas not in options:
                    msg = 'class {} not available for band {}, only {}'
                    raise ValueError(msg.format(clas, mask_band, options))

        if classes == 'all' or classes == ['all']:
            selection = options
        else:
            selection = classes

        mask = self.bitImage(image, mask_band, renamed).select(selection)
        return mask

    def applyMask(self, image, mask_band, classes='all', renamed=False):
        """ Apply a mask """
        # SCL Mask
        if classes == 'all':
            if self.id == 'COPERNICUS/S2_SR' and mask_band == 'SCL':
                bands = ['dark','shadow', 'cloud_low', 'cloud_medium',
                         'cloud_high', 'cirrus_cloud', 'snow']
            else:
                bands = None
        else:
            bands = classes

        if isinstance(mask_band, str):
            mask_band = self.getMask(image, mask_band, classes, renamed)

        return tools.image.applyMask(image, mask_band, bands)

    def applyPositiveMask(self, image, mask_band, classes='all',
                          renamed=False):
        """ Apply a positive mask (for example: clear) """
        mask = self.getMask(image, mask_band, classes, renamed)
        options = mask.bandNames()

        def wrap(band, img):
            img = ee.Image(img)
            band = ee.String(band)
            m = mask.select(band)
            return img.updateMask(m)

        return ee.Image(options.iterate(wrap, image))

    def getBand(self, band, by='id'):
        """ get a band by its id or name """
        data = None
        for b in self.bands:
            if by == 'name':
                bid = b.name
            else:
                bid = b.id

            if bid == band:
                data = b

        return data

    def visualization(self, colors, renamed=False):
        """ Return visualization parameters for ui.Map.addLayer.

        :param colors: 'NSR' (nir swir red), 'NSR2' ((nir swir2 red),
            'RGB' (red green blue), 'falseColor' (nir red green)
        """
        options = ['NSR', 'NSR2', 'RGB', 'falseColor', 'SCL']
        vis = {}
        b = self.getBand('blue', 'name')
        g = self.getBand('green', 'name')
        r = self.getBand('red', 'name')
        n = self.getBand('nir', 'name')
        s = self.getBand('swir', 'name')
        s2 = self.getBand('swir2', 'name')
        scl = self.getBand('scene_classification_map', 'name')

        def register(one, two, three, factor, name):
            if renamed:
                bandone = one.name
                bandtwo = two.name
                bandthree = three.name
            else:
                bandone = one.id
                bandtwo = two.id
                bandthree = three.id

            vis[name] = {
                'bands': [bandone, bandtwo, bandthree],
                'min': [one.min, two.min, three.min],
                'max': [one.max/factor, two.max/factor, three.max/factor]
            }

        if n and s2 and r:
            register(n, s2, r, 2, 'NSR2')

        if n and s and r:
            register(n, s, r, 2, 'NSR')

        if r and g and b:
            register(r, g, b, 3, 'RGB')

        if n and r and g:
            register(n, r, g, 2, 'falseColor')

        if scl:
            if renamed:
                band = scl.name
            else:
                band = scl.id

            vis['SCL'] = {
                'bands': [band],
                'min': scl.min,
                'max': scl.max,
                'palette': ['ff0004', '868686', '774b0a', '10d22c',
                            'ffff52', '0000ff', '818181', 'c0c0c0',
                            'f1f1f1', 'bac5eb', '52fff9']
            }

        if colors in options:
            return vis[colors]
        else:
            return {}

    def ndvi(self, image, name='ndvi', renamed=False):
        if renamed:
            n = 'nir'
            r = 'red'
        else:
            n = self.getBand('nir', 'name').id
            r = self.getBand('red', 'name').id
        if n and r:
            return indices.ndvi(image, n, r, name)
        else:
            raise ValueError('ndvi index cannot be computed in {}'.format(
                self.id
            ))

    def evi(self, image, name='evi', renamed=False):
        if renamed:
            n = 'nir'
            r = 'red'
            b = 'blue'
        else:
            n = self.getBand('nir', 'name').id
            r = self.getBand('red', 'name').id
            b = self.getBand('blue', 'name').id
        if n and r and b:
            return indices.evi(image, n, r, b, bandname=name)
        else:
            raise ValueError('evi index cannot be computed in {}'.format(
                self.id
            ))

    def nbr(self, image, name='nbr', renamed=False):
        if renamed:
            n = 'nir'
            s = 'swir2'
            if not s:
                s = 'swir'
        else:
            n = self.getBand('nir', 'name').id
            s = self.getBand('swir2', 'name').id
            if not s:
                s = self.getBand('swir', 'name').id
        if n and s:
            return indices.nbr(image, n, s, name)
        else:
            raise ValueError('nbr index cannot be computed in {}'.format(
                self.id
            ))

    # common indices
    indices = {
        'ndvi': ndvi,
        'evi': evi,
        'nbr': nbr
    }

    def bitImage(self, image, qa, renamed=False):
        """ Get an image from the bit information from the qa band

        :param qa: the quality band name
        :type qa: str
        :param image: the image to decode with the qa band
        :type image: ee.Image
        :return: the image with the decode bands added
        """
        if renamed:
            band = self.getBand(qa, 'name')
        else:
            band = self.getBand(qa, 'id')

        if not band:
            raise ValueError('{} band not present in {}'.format(
                qa, self.id
            ))

        reader = bitreader.BitReader(band.bits)
        return reader.decodeImage(image, qa)

    def checkBands(self, bands):
        """ Check if all bands have the same max and min

        :type collection: Collection
        :type bands: list
        :rtype: bool
        """
        bandsmax = []
        bandsmin = []
        for band in bands:
            bandmax = band.max
            bandmin = band.min
            bandsmax.append(bandmax)
            bandsmin.append(bandmin)

        if not allequal(bandsmax) or not allequal(bandsmin):
            return False
        return True

    def rename(self, image, reference='all'):
        """ Rename bands according to the parsed reference. It can be:
        optical, thermal, bits, all """
        if reference == 'all':
            original_names = {band.id: band.name for band in self.bands}
        else:
            original_names = {band.id: band.name for band in self.bands if band.reference == reference}

        return tools.image.renameDict(image, original_names)

    def proxyImage(self, renamed=False):
        """ Create an Image with the band names, type and scale but empty """
        precisions = self.precisions(renamed=renamed)
        first_band = self.bands[0]
        if not renamed:
            name = first_band.id
        else:
            name = first_band.name

        init = ee.Image.constant(0).rename(name)
        init = convertPrecision(init, precisions[name])
        for i, band in enumerate(self.bands):
            if i == 0: continue
            if not renamed:
                name = band.id
            else:
                name = band.name

            img = ee.Image.constant(0).rename(name)
            img = convertPrecision(img, precisions[name])
            init = init.addBands(img)

        return init


from .landsat import *
from .sentinel import *
from . import landsat, sentinel, modis
from .landsat import Landsat
from .sentinel import Sentinel2
from .modis import MODIS

IDS = landsat.IDS + sentinel.IDS + modis.IDS


def fromId(id):
    """ Create a collection from a parsed ID """
    functions = [
        landsat.Landsat.fromId,
        sentinel.Sentinel2.fromId,
        modis.MODIS.fromId
    ]
    for f in functions:
        try:
            col = f(id)
        except:
            continue
        else:
            return col

    # tried all collections and did not find it
    raise ValueError('{} not recognized as a valid ID'.format(id))


def fromShortName(short_name):
    """ Create a collection from a short name """
    functions = [
        landsat.Landsat.fromShortName,
        sentinel.Sentinel2.fromShortName,
        modis.MODIS.fromShortName
    ]
    for f in functions:
        try:
            col = f(short_name)
        except:
            continue
        else:
            return col

    # tried all collections and did not find it
    raise ValueError('{} not recognized as a valid short name'.format(short_name))


def getCommonBands(*collections, reference='all', match='id'):
    """ Get the common bands of the parsed collections

    :param match: the field to match, can be: id or name
    :type match: str
    """
    first = collections[0]
    renamed = True if match == 'name' else False

    if len(collections) == 1:
        return first.bandNames(reference, renamed)

    first_set = set(first.bandNames(reference, renamed))
    if len(collections) == 1:
        return first.bandNames
    else:
        rest = collections[1:]
        for col in rest:
            bandset = set(col.bandNames(reference, renamed))
            first_set = first_set.intersection(bandset)

    return list(first_set)


def rescale(image, col, collection_to_match, reference='all', renamed=False,
            drop=False):
    """ Re-scale the values of image which must belong to collection so the
        values match the ones from collection_from

    :param collection: The Collection to which belongs the image
    :type collection: Collection
    :param collection_to_match: the Collection to get the range from
    :type collection_to_match: Collection
    :param reference: optical, thermal, bits or all
    :type reference: str
    """
    # Create comparative collection
    # bands = ee.Dictionary(col.bands)
    common_bands = getCommonBands(col, collection_to_match,
                                  reference=reference, match='name')
    # keep only bands with min and max values
    new_common = []

    ranges = {}
    ranges_other = {}

    def setrange(band, range_dict):
        if not renamed:
            name = band.id
        else:
            name = band.name

        range_dict[name] = {'min': band.min, 'max': band.max}

    precisions = {}
    for band in common_bands:
        b = col.getBand(band, 'name')
        b_proxy = collection_to_match.getBand(band, 'name')
        if b.min is not None and \
                b.max is not None and \
                b_proxy.min is not None and \
                b_proxy.max is not None:
            if not renamed:
                name = b.id
            else:
                name = b.name
            new_common.append(name)
            setrange(b, ranges)
            setrange(b_proxy, ranges_other)
            precisions[name] = b_proxy.precision

    new_common = ee.List(new_common)
    ranges_this = ee.Dictionary(ranges)
    ranges_proxy = ee.Dictionary(ranges_other)
    precisions = ee.Dictionary(precisions)

    def iteration(band, ini):
        ini = ee.Image(ini)
        band = ee.String(band)
        ranges_this_band = ee.Dictionary(ranges_this.get(band))
        ranges_proxy_band = ee.Dictionary(ranges_proxy.get(band))
        min_this = ee.Number(ranges_this_band.get('min'))
        min_proxy = ee.Number(ranges_proxy_band.get('min'))
        max_this = ee.Number(ranges_this_band.get('max'))
        max_proxy = ee.Number(ranges_proxy_band.get('max'))

        equal_min = min_this.eq(min_proxy)
        equal_max = max_this.eq(max_proxy)
        equal = equal_min.And(equal_max)

        def true(ini):
            return ini

        def false(ini, band, min_this, max_this, min_proxy, max_proxy):
            return tools.image.parametrize(ini,
                                           (min_this, max_this),
                                           (min_proxy, max_proxy),
                                           bands=[band])

        return ee.Image(ee.Algorithms.If(
            equal, true(ini),
            false(ini, band, min_this, max_this, min_proxy, max_proxy)))

    final = ee.Image(new_common.iterate(iteration, image))
    final = convertPrecisions(final, precisions)
    if drop:
        final = final.select(new_common)

    return final

from .group import CollectionGroup

# Preload factory methods
Landsat1 = Landsat.Landsat1
Landsat2 = Landsat.Landsat2
Landsat3 = Landsat.Landsat3
Landsat4SR = Landsat.Landsat4SR
Landsat4TOA = Landsat.Landsat4TOA
Landsat5SR = Landsat.Landsat5SR
Landsat5TOA = Landsat.Landsat5TOA
Landsat7SR = Landsat.Landsat7SR
Landsat7TOA = Landsat.Landsat7TOA
Landsat8SR = Landsat.Landsat8SR
Landsat8TOA = Landsat.Landsat8TOA
Sentinel2TOA = Sentinel2.Sentinel2TOA
Sentinel2SR = Sentinel2.Sentienl2SR
MOD09GQ = MODIS.MOD09GQ
MYD09GQ = MODIS.MYD09GQ
MOD09GA = MODIS.MOD09GA
MYD09GA = MODIS.MYD09GA
MOD13Q1 = MODIS.MOD13Q1
MYD13Q1 = MODIS.MYD13Q1